unit adxOLCustomizeNewButton_IMPL;

interface

uses
  SysUtils, ComObj, ComServ, ActiveX, Variants, adxAddIn,
  adxOLCustomizeNewButton_TLB, StdVcl, Classes,
  Office2000, Outlook2000, Dialogs;

type
  TADX_OLCustomizeNewButton = class(TadxAddin, IADX_OLCustomizeNewButton)
  end;

  TAddInModule = class(TadxCOMAddInModule)
    procedure adxCOMAddInModuleAddInStartupComplete(Sender: TObject);
    procedure adxCOMAddInModuleAddInFinalize(Sender: TObject);
  private
    FEventSink: Pointer;
    FButton: CommandBarButton;
  protected
  public
    procedure DoClick(AControl: CommandBarControl);
  end;

implementation

{$R *.dfm}

uses Windows;

type
  IButtonEvents = interface(IDispatch)
    ['{000C0351-0000-0000-C000-000000000046}']
  end;

  TButtonEventSink = class(TInterfacedObject, IButtonEvents)
  private
    FAddInModule: TAddInModule;
    FIControl: CommandBarButton;
    FCookie: Longint;
    { IDispatch }
    function GetTypeInfoCount(out Count: Integer): HResult; stdcall;
    function GetTypeInfo(Index, LocaleID: Integer; out TypeInfo): HResult; stdcall;
    function GetIDsOfNames(const IID: TGUID; Names: Pointer;
      NameCount, LocaleID: Integer; DispIDs: Pointer): HResult; stdcall;
    function Invoke(DispID: Integer; const IID: TGUID; LocaleID: Integer;
      Flags: Word; var Params; VarResult, ExcepInfo, ArgErr: Pointer): HResult; stdcall;
  protected
    procedure Disconnect;
  public
    constructor Create(AddInModule: TAddInModule; AControl: CommandBarButton);
    destructor Destroy; override;
  end;

{ TButtonEventSink }

constructor TButtonEventSink.Create(AddInModule: TAddInModule; AControl: CommandBarButton);
begin
  inherited Create;
  FAddInModule := AddInModule;
  FIControl := AControl;
  InterfaceConnect(AControl, DIID__CommandBarButtonEvents, Self, FCookie)
end;

destructor TButtonEventSink.Destroy;
begin
  FIControl := nil;
  FAddInModule := nil;
  inherited Destroy;
end;

procedure TButtonEventSink.Disconnect;
begin
  try
    if Assigned(FIControl) then
      InterfaceDisconnect(FIControl, DIID__CommandBarButtonEvents, FCookie);
  except
    // Outlook 2000 bug
  end;
end;

function TButtonEventSink.GetTypeInfoCount(out Count: Integer): HResult;
begin
  Count := 0;
  Result := S_OK;
end;

function TButtonEventSink.GetTypeInfo(Index, LocaleID: Integer; out TypeInfo): HResult;
begin
  Pointer(TypeInfo) := nil;
  Result := E_FAIL;
end;

function TButtonEventSink.GetIDsOfNames(const IID: TGUID; Names: Pointer;
  NameCount, LocaleID: Integer; DispIDs: Pointer): HResult;
var
  s: string;
  i: integer;
begin
  Result := S_OK;
  if DispIDs = nil then Result := E_POINTER;
  if (Result = S_OK) and (Names = nil) then Result := E_POINTER;
  if (Result = S_OK) and (NameCount <> 1) then Result := E_INVALIDARG;
  if Result = S_OK then begin
    i := 512;
    SetLength(s, i);
    try
      WideCharToMultiByte(CP_ACP, 0, Names, -1, PChar(s), i, nil, nil);
      case FIControl.Type_ of
        msoControlButton:
          i := lstrcmp(PChar(s), 'Click');
        msoControlEdit, msoControlComboBox, msoControlDropdown:
          i := lstrcmp(PChar(s), 'Change');
      end;
      if i <> 0 then Result := DISP_E_UNKNOWNNAME
      else PInteger(DispIDs)^ := $00000001;
    finally
      SetLength(s, 0);
    end;
  end;
end;

function TButtonEventSink.Invoke(DispID: Integer; const IID: TGUID;
  LocaleID: Integer; Flags: Word; var Params; VarResult, ExcepInfo,
  ArgErr: Pointer): HResult;
type
  TVariantArray = array of OleVariant;
var
  FTag: WideString;
  ParamPtr: ^Integer;
  ParamCount, i: Integer;
  VarArray: TVariantArray;
  IDsp: IDispatch;
  Ctrl: CommandBarControl;
begin
  Result := S_OK;
  ParamPtr := @Params;
  if DispID <> $00000001 then Result := DISP_E_MEMBERNOTFOUND;
  if (Result = S_OK) and (ParamPtr = nil) then Result := E_POINTER;
  if Result = S_OK then
    try
      ParamCount := TDispParams(Params).cArgs;
      SetLength(VarArray, ParamCount);
      for i := Low(VarArray) to High(VarArray) do
        VarArray[High(VarArray) - i] := OleVariant(TDispParams(Params).rgvarg^[i]);
      IDsp := VarArray[0];
      IDsp.QueryInterface(CommandBarControl, Ctrl);
      FTag := '';
      if Assigned(FIControl) then FTag := FIControl.Tag;
      if Ctrl.Tag = FTag then
        try
          FAddInModule.DoClick(FIControl);
        except
        end;
    finally
      Ctrl := nil;
      IDsp := nil;
      SetLength(VarArray, 0);
    end;
end;

{ TAddInModule }

procedure TAddInModule.adxCOMAddInModuleAddInStartupComplete(Sender: TObject);
var
  IPopup: CommandBarPopup;
  ICmdBar: CommandBar;
begin
  // Parent = CommandBars['Standard']
  // ID = 1874
  // Type = $000D (???)
  if (OutlookApp.ActiveExplorer <> nil) then begin
    ICmdBar := OutlookApp.ActiveExplorer.CommandBars.Item['Standard'];
    if Assigned(ICmdBar) then
      try
        IPopup := ICmdBar.FindControl(EmptyParam, 1874, EmptyParam, EmptyParam, True) as CommandBarPopup;
        if Assigned(IPopup) then
          try
            FButton := IPopup.Controls.Add(msoControlButton, EmptyParam, EmptyParam, EmptyParam, True) as CommandBarButton;
            if Assigned(FButton) then begin
              // unique ID
              FButton.Set_Tag('{1B8CAF80-2C18-4FEB-8C38-79D4D0D211A6}');
              FButton.Set_Caption('My Caption');
              FButton.Set_TooltipText('My Caption');
              FButton.Set_Visible(True);
              FButton.Set_Enabled(True);

              FEventSink := TButtonEventSink.Create(Self, FButton);
              //
              // TODO
              //
            end;
          finally
            IPopup := nil;
          end;
      finally
        ICmdBar := nil;
      end;
  end;
end;

procedure TAddInModule.adxCOMAddInModuleAddInFinalize(Sender: TObject);
begin
  if Assigned(FButton) then begin
    TButtonEventSink(FEventSink).Disconnect;
    FEventSink := nil;
    FButton := nil;
  end;
end;

procedure TAddInModule.DoClick(AControl: CommandBarControl);
begin
  ShowMessage('Click');
end;

initialization
  TadxFactory.Create(ComServer, TADX_OLCustomizeNewButton, CLASS_ADX_OLCustomizeNewButton, TAddInModule);

end.
